import os
import re
import glob
import json
import logging
from datetime import datetime
from contextlib import contextmanager
from crmsh import utils as crmshutils
from . import config
from crmsh import ra
from crmsh import log
from crmsh.sh import ShellUtils
from crmsh import constants


logger = log.setup_logger(__name__)


CRED = constants.RED
CYELLOW = constants.YELLOW
CGREEN = constants.GREEN
CEND = constants.END

LEVEL = {
    "info": logging.INFO,
    "warn": logging.WARNING,
    "error": logging.ERROR
}


class MyLoggingFormatter(logging.Formatter):
    """
    Class to change logging formatter
    """

    FORMAT_FLUSH = "[%(asctime)s]%(levelname)s: %(message)s"
    FORMAT_NOFLUSH = "%(timestamp)s%(levelname)s: %(message)s"

    COLORS = {
        'WARNING': CYELLOW,
        'INFO': CGREEN,
        'ERROR': CRED
    }

    def __init__(self, flush=True):
        fmt = self.FORMAT_FLUSH if flush else self.FORMAT_NOFLUSH
        logging.Formatter.__init__(self, fmt=fmt, datefmt='%Y/%m/%d %H:%M:%S')

    def format(self, record):
        levelname = record.levelname
        if levelname in self.COLORS:
            levelname_color = self.COLORS[levelname] + levelname + CEND
            record.levelname = levelname_color
        return logging.Formatter.format(self, record)


def now(form="%Y/%m/%d %H:%M:%S"):
    return datetime.now().strftime(form)


@contextmanager
def manage_handler(_type, keep=True):
    """
    Define a contextmanager to remove specific logging handler temporarily
    """
    try:
        handler = get_handler(logger, _type)
        if not keep:
            logger.removeHandler(handler)
        yield
    finally:
        if not keep:
            logger.addHandler(handler)


def msg_raw(level, msg, to_stdout=True):
    with manage_handler("console", to_stdout):
        logger.log(level, msg)


def msg_info(msg, to_stdout=True):
    msg_raw(logging.INFO, msg, to_stdout)


def msg_warn(msg, to_stdout=True):
    msg_raw(logging.WARNING, msg, to_stdout)


def msg_error(msg, to_stdout=True):
    msg_raw(logging.ERROR, msg, to_stdout)


def json_dumps():
    """
    Dump the json results to file
    """
    from . import main
    with open(main.ctx.jsonfile, 'w') as f:
        f.write(json.dumps(main.ctx.task_list, indent=2))
        f.flush()
        os.fsync(f)


class FenceInfo(object):
    """
    Class to collect fence info
    """
    @property
    def fence_enabled(self):
        enable_result = crmshutils.get_property("stonith-enabled")
        if not enable_result or enable_result.lower() != "true":
            return False
        return True

    @property
    def fence_configured(self):
        return crmshutils.has_stonith_running()

    @property
    def fence_action(self):
        action_result = crmshutils.get_property("stonith-action")
        supported_actions = ra.get_property_options("stonith-action")
        if action_result is None or action_result not in supported_actions:
            msg_error(f"Cluster property \"stonith-action\" should be {'|'.join(supported_actions)}")
            return None
        return action_result

    @property
    def fence_timeout(self):
        timeout_result = crmshutils.get_property("stonith-timeout")
        if timeout_result and re.match(r'[1-9][0-9]*(s|)$', timeout_result):
            return timeout_result.strip("s")
        return config.FENCE_TIMEOUT


def check_node_status(node, state):
    """
    Check whether the node has expected state
    """
    rc, stdout, stderr = ShellUtils().get_stdout_stderr('crm_node -l')
    if rc != 0:
        msg_error(stderr)
        return False
    pattern = re.compile(r'^.* {} {}'.format(node, state), re.MULTILINE)
    if not pattern.search(stdout):
        return False
    return True


def online_nodes():
    """
    Get online node list
    """
    rc, stdout, stderr = ShellUtils().get_stdout_stderr('crm_mon -1')
    if rc == 0 and stdout:
        res = re.search(r'Online:\s+\[\s(.*)\s\]', stdout)
        if res:
            return res.group(1).split()
    return []


def peer_node_list():
    """
    Get online node list except self
    """
    online_nodelist = online_nodes()
    if online_nodelist:
        online_nodelist.remove(this_node())
        return online_nodelist
    return []


def this_node():
    """
    Try to get the node name from crm_node command
    If failed, use its hostname
    """
    rc, stdout, stderr = ShellUtils().get_stdout_stderr("crm_node --name")
    if rc != 0:
        msg_error(stderr)
        return crmshutils.this_node()
    return stdout


def str_to_datetime(str_time, fmt):
    return datetime.strptime(str_time, fmt)


def corosync_port_list():
    """
    Get corosync ports using corosync-cmapctl
    """
    ports = []
    rc, out, _ = ShellUtils().get_stdout_stderr("corosync-cmapctl totem.interface")
    if rc == 0 and out:
        ports = re.findall(r'(?:mcastport.*) ([0-9]+)', out)
    return ports


def get_handler(logger, _type):
    """
    Get logger specific handler
    """
    for h in logger.handlers:
        if getattr(h, '_name') == _type:
            return h


def is_root():
    return os.getuid() == 0


def get_process_status(s):
    """
    Returns true if argument is the name of a running process.

    s: process name
    returns Boolean and pid
    """
    # find pids of running processes
    pids = [pid for pid in os.listdir('/proc') if pid.isdigit()]
    for pid in pids:
        try:
            pid_file = os.path.join('/proc', pid, 'cmdline')
            with open(pid_file, 'rb') as f:
                data = f.read()
                procname = os.path.basename(crmshutils.to_ascii(data).replace('\x00', ' ').split(' ')[0])
                if procname == s or procname == s + ':':
                    return True, int(pid)
        except EnvironmentError:
            # a process may have died since we got the list of pids
            pass
    return False, -1


def _find_match_count(str1, str2):
    """
    Find the max match number of s1 and s2
    """
    leng = min(len(str1), len(str2))
    num = 0

    for i in range(leng):
        if str1[i] == str2[i]:
            num += 1
        else:
            break

    return num


def is_valid_sbd(dev):
    """
    Check whether the device is a initialized SBD device

    dev: dev path
    return 'True' if 'dev' is a initialized SBD device
    """
    if not os.path.exists(dev):
        return False

    rc, out, err = ShellUtils().get_stdout_stderr(config.SBD_CHECK_CMD.format(dev=dev))
    if rc != 0 and err:
        msg_error(err)
        return False

    return True


def find_candidate_sbd(dev):
    """
    Find the devices that already have SBD header

    return the path of candidate SBD device
    """
    ddir = os.path.dirname(dev)
    dname = os.path.basename(dev)

    dev_list = glob.glob(ddir + "/*")
    if len(dev_list) == 0:
        return ""

    can_filter = filter(is_valid_sbd, dev_list)
    candidates = list(can_filter)

    if len(candidates) == 0:
        return ""

    index = 0
    i = 0
    max_match = -1
    num_list = map(_find_match_count, [dev] * len(candidates), candidates)
    for num in num_list:
        if num > max_match:
            max_match = num
            index = i
        i += 1

    return candidates[index]


def warning_ask(warn_string):
    from . import main
    if main.ctx.force:
        return False
        
    try:
        ans = input(CYELLOW + warn_string + CEND)
    except EOFError:
        return False
    return True if ans == "Yes" else False



# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# See COPYING for license information.

import os
import sys
from .utils import rmdir_r, quote, this_node, ext_cmd
from .xmlutil import get_topmost_rsc, get_op_timeout, get_child_nvset_node, is_ms_or_promotable_clone, is_cloned
from . import log


logger = log.setup_logger(__name__)

#
# Resource testing suite
#


class RADriver(object):
    '''
    Execute operations on resources.
    '''
    pfx = {
        "instance_attributes": "OCF_RESKEY_",
        "meta_attributes": "OCF_RESKEY_CRM_meta_",
    }
    undef = -200
    unused = -201

    def __init__(self, rsc_node, nodes):
        from tempfile import mkdtemp
        self.rscdef_node = rsc_node
        if rsc_node is not None:
            self.ra_class = rsc_node.get("class")
            self.ra_type = rsc_node.get("type")
            self.ra_provider = rsc_node.get("provider")
            self.ident = rsc_node.get("id")
        else:
            self.ra_class = None
            self.ra_type = None
            self.ra_provider = None
            self.ident = None
        self.nodes = nodes
        self.outdir = mkdtemp(prefix="crmsh_out.")
        self.errdir = mkdtemp(prefix="crmsh_err.")
        self.ec_l = {}
        self.ec_ok = self.unused
        self.ec_stopped = self.unused
        self.ec_promoted = self.unused
        self.last_op = None
        self.last_rec = {}
        self.timeout = 20000

    def __del__(self):
        rmdir_r(self.outdir)
        rmdir_r(self.errdir)

    def id_str(self):
        return self.last_op and "%s:%s" % (self.ident, self.last_op) or self.ident

    def err(self, s):
        logger.error("%s: %s", self.id_str(), s)

    def warn(self, s):
        logger.warning("%s: %s", self.id_str(), s)

    def info(self, s):
        logger.info("%s: %s", self.id_str(), s)

    def debug(self, s):
        logger.debug("%s: %s", self.id_str(), s)

    def is_ms_or_promotable_clone(self):
        return is_ms_or_promotable_clone(get_topmost_rsc(self.rscdef_node))

    def nvset2env(self, set_n):
        if set_n is None:
            return
        try:
            pfx = self.pfx[set_n.tag]
        except:
            self.err("unknown attributes set: %s" % set_n.tag)
            return
        for nvpair in set_n.iterchildren():
            if nvpair.tag != "nvpair":
                continue
            n = nvpair.get("name")
            v = nvpair.get("value")
            self.rscenv["%s%s" % (pfx, n)] = v

    def set_rscenv(self, op):
        '''
        Setup the environment. Class specific.
        '''
        self.rscenv = {}
        n = self.rscdef_node
        self.timeout = get_op_timeout(n, op, "20s")
        self.rscenv["%stimeout" % self.pfx["meta_attributes"]] = str(self.timeout)
        if op == "monitor":
            self.rscenv["%sinterval" % self.pfx["meta_attributes"]] = "10000"
        if is_cloned(n):
            # some of the meta attributes for clones/ms are used
            # by resource agents
            cn = get_topmost_rsc(n)
            self.nvset2env(get_child_nvset_node(cn))

    def op_status(self, host):
        'Status of the last op.'
        try:
            return self.ec_l[host]
        except:
            return self.undef

    def explain_op_status(self, host):
        stat = self.op_status(host)
        if stat == -9:
            return "timed out"
        elif stat == self.undef:
            return "unknown reason (the RA couldn't run?)"
        else:
            return "exit code %d" % stat

    def is_ok(self, host):
        'Was last op successful?'
        return self.op_status(host) == self.ec_ok

    def is_promoted(self, host):
        'Only if last op was probe/monitor.'
        return self.op_status(host) == self.ec_promoted

    def is_stopped(self, host):
        'Only if last op was probe/monitor.'
        return self.op_status(host) == self.ec_stopped

    def show_log(self, host):
        '''
        Execute an operation.
        '''
        try:
            from .crm_pssh import show_output
        except ImportError:
            logger.error("Parallax SSH not installed, rsctest can not be executed")
            return

        logger.error("host %s (%s)\n" %
                         (host, self.explain_op_status(host)))
        show_output(self.errdir, [host], "stderr")
        show_output(self.outdir, [host], "stdout")

    def run_on_all(self, op):
        '''
        In case of cloned resources, it doesn't make sense to run
        (certain) operations on just one node. So, we run them
        everywhere instead.
        For instance, some clones require quorum.
        '''
        return is_cloned(self.rscdef_node) and op in ("start", "stop")

    def exec_cmd(self, op):
        '''defined in subclasses'''
        pass

    def runop(self, op, nodes=None, local_only=False):
        '''
        Execute an operation.
        '''
        if not nodes or self.run_on_all(op):
            nodes = self.nodes
        self.last_op = op
        self.set_rscenv(op)
        real_op = (op == "probe" and "monitor" or op)
        cmd = self.exec_cmd(real_op)
        logger.debug("running %s on %s", real_op, nodes)
        for attr in self.rscenv:
            # shell doesn't allow "-" in var names
            envvar = attr.replace("-", "_")
            cmd = "%s=%s %s" % (envvar, quote(self.rscenv[attr]), cmd)
        if local_only:
            self.ec_l[this_node()] = ext_cmd(cmd)
        else:
            try:
                from .crm_pssh import do_pssh_cmd
            except ImportError:
                logger.error("Parallax SSH not installed, rsctest can not be executed")
                return

            results = do_pssh_cmd(cmd, nodes, self.outdir, self.errdir, self.timeout)
            for i, node in enumerate(nodes):
                try:
                    host, result = results[i]
                except IndexError:
                    self.ec_l[node] = self.undef
                else:
                    assert node == host
                    self.ec_l[node] = result.returncode
        return

    def stop(self, node):
        """
        Make sure resource is stopped on node.
        """
        if self.is_ms_or_promotable_clone():
            self.runop("demote", (node,))
        self.runop("stop", (node,))
        ok = self.is_ok(node)
        if not ok:
            self.err("resource failed to stop on %s, clean it up!" % node)
            self.show_log(node)
        return ok

    def test_resource(self, node):
        """
        Perform test of resource on node.
        """
        self.runop("start", (node,))
        if self.is_ms_or_promotable_clone() and self.is_ok(node):
            self.runop("promote", (node,))
        return self.is_ok(node)

    def probe(self):
        """
        Execute probe (if possible)
        """
        self.runop("probe")

    def verify_stopped(self, node):
        """
        Make sure resource is stopped on node.
        """
        stopped = self.is_stopped(node)
        if not stopped:
            if self.is_ok(node):
                self.warn("resource running at %s" % (node))
            elif self.is_ms_or_promotable_clone() and self.is_promoted(node):
                self.warn("resource is promoted at %s" % (node))
            else:
                self.warn("resource not clean at %s" % (node))
                self.show_log(node)
        return stopped


class RAOCF(RADriver):
    '''
    Execute operations on OCF resources.
    '''
    # OCF exit codes
    OCF_SUCCESS = 0
    OCF_ERR_GENERIC = 1
    OCF_ERR_ARGS = 2
    OCF_ERR_UNIMPLEMENTED = 3
    OCF_ERR_PERM = 4
    OCF_ERR_INSTALLED = 5
    OCF_ERR_CONFIGURED = 6
    OCF_NOT_RUNNING = 7
    OCF_RUNNING_PROMOTED = 8
    OCF_FAILED_PROMOTED = 9

    def __init__(self, *args):
        RADriver.__init__(self, *args)
        self.ec_ok = self.OCF_SUCCESS
        self.ec_stopped = self.OCF_NOT_RUNNING
        self.ec_promoted = self.OCF_RUNNING_PROMOTED

    def set_rscenv(self, op):
        RADriver.set_rscenv(self, op)
        self.nvset2env(get_child_nvset_node(self.rscdef_node, "instance_attributes"))
        self.rscenv["OCF_RESOURCE_INSTANCE"] = self.ident
        self.rscenv["OCF_ROOT"] = os.environ["OCF_ROOT"]

    def exec_cmd(self, op):
        cmd = "%s/resource.d/%s/%s %s" % \
            (os.environ["OCF_ROOT"], self.ra_provider, self.ra_type, op)
        return cmd


class RALSB(RADriver):
    '''
    Execute operations on LSB resources (init scripts).
    '''

    # OCF exit codes
    LSB_OK = 0
    LSB_ERR_GENERIC = 1
    LSB_ERR_ARGS = 2
    LSB_ERR_UNIMPLEMENTED = 3
    LSB_ERR_PERM = 4
    LSB_ERR_INSTALLED = 5
    LSB_ERR_CONFIGURED = 6
    LSB_NOT_RUNNING = 7
    LSB_STATUS_DEAD_PID = 1
    LSB_STATUS_DEAD_LOCK = 2
    LSB_STATUS_NOT_RUNNING = 3
    LSB_STATUS_UNKNOWN = 4

    def __init__(self, *args):
        RADriver.__init__(self, *args)
        self.ec_ok = self.LSB_OK
        self.ec_stopped = self.LSB_STATUS_NOT_RUNNING
        self.ec_promoted = self.unused

    def set_rscenv(self, op):
        RADriver.set_rscenv(self, op)

    def exec_cmd(self, op):
        if self.ra_type.startswith("/"):
            prog = self.ra_type
        else:
            prog = "/etc/init.d/%s" % self.ra_type
        cmd = "%s %s" % (prog, op == "monitor" and "status" or op)
        return cmd


class RASystemd(RADriver):
    '''
    Execute operations on systemd resources.
    '''

    # Error codes are meaningless for systemd...
    SYSD_OK = 0
    SYSD_ERR_GENERIC = 1
    SYSD_NOT_RUNNING = 3

    def __init__(self, *args):
        RADriver.__init__(self, *args)
        self.ec_ok = self.SYSD_OK
        self.ec_stopped = self.SYSD_NOT_RUNNING
        self.ec_promoted = self.unused

    def set_rscenv(self, op):
        RADriver.set_rscenv(self, op)

    def exec_cmd(self, op):
        op = "status" if op == "monitor" else op
        cmd = "systemctl %s %s.service" % (op, self.ra_type)
        return cmd


class RAStonith(RADriver):
    '''
    Execute operations on Stonith resources.
    '''

    STONITH_OK = 0
    STONITH_ERR = 1

    def __init__(self, *args):
        RADriver.__init__(self, *args)
        self.ec_ok = self.STONITH_OK
        self.ec_stopped = self.STONITH_ERR

    def stop(self, node):
        """
        Disable for stonith resources.
        """
        return True

    def verify_stopped(self, node):
        """
        Disable for stonith resources.
        """
        return True

    def test_resource(self, node):
        """
        Run test for stonith resource
        """
        for prefix in ['fence_', ]:
            if self.ra_type.startswith(prefix):
                self.err("Cannot test STONITH resources!")
                return False
        return RADriver.test_resource(self, node)

    def set_rscenv(self, op):
        RADriver.set_rscenv(self, op)
        for nv in self.rscdef_node.xpath("instance_attributes/nvpair"):
            self.rscenv[nv.get('name')] = nv.get('value')

    def exec_cmd(self, op):
        """
        Probe resource on each node.
        """
        return "stonith -t %s -E -S" % (self.ra_type)


ra_driver = {
    "ocf": RAOCF,
    "lsb": RALSB,
    "stonith": RAStonith,
    "systemd": RASystemd
}


def check_test_support(resources):
    rc = True
    for r in resources:
        ra_class = r.get("class")
        if not ra_class:
            logger.warning("class attribute not found in %s", r.get('id'))
            rc = False
        elif ra_class not in ra_driver:
            logger.warning("testing of class %s resources not supported", ra_class)
            rc = False
    return rc


def are_all_stopped(resources, nodes):
    rc = True
    logger.info("Probing resources")
    for r in resources:
        ra_class = r.get("class")
        drv = ra_driver[ra_class](r, nodes)
        drv.probe()
        for node in nodes:
            if not drv.verify_stopped(node):
                rc = False
    return rc


def stop_all(started, node):
    'Stop all started resources in reverse order on node.'
    while started:
        drv = started.pop()
        drv.stop(node)


def test_resources(resources, nodes, all_nodes):
    def test_node(node):
        started = []
        id_list_str = ' '.join([r.get("id") for r in resources])
        logger.info(f"Testing on {node}: {id_list_str}")
        for r in resources:
            ra_class = r.get("class")
            drv = ra_driver[ra_class](r, (node,))
            if drv.test_resource(node):
                started.append(drv)
            else:
                drv.show_log(node)
                stop_all(started, node)
                return False
        stop_all(started, node)
        return True

    if not check_test_support(resources):
        return False
    if not are_all_stopped(resources, all_nodes):
        logger.error("Stop all resources before testing!\n")
        return False
    rc = True
    for node in nodes:
        rc |= test_node(node)
    return rc


def call_resource(rsc, cmd, nodes, local_only):
    """
    Calls the given operation on the resource.
    local_only: Only performs the call locally (don't use SSH).
    """
    ra_class = rsc.get("class")
    if ra_class not in ra_driver:
        logger.error("Calling '%s' for resource not supported", cmd)
        return False
    d = ra_driver[ra_class](rsc, [])

    from . import ra
    agent = ra.get_ra(rsc)
    actions = list(agent.actions().keys()) + ['meta-data', 'validate-all']

    if cmd not in actions:
        logger.error("action '%s' not supported by %s", cmd, agent)
        return False
    d.runop(cmd, nodes, local_only=local_only)
    for node in nodes:
        ok = d.is_ok(node)
        if not ok:
            logger.error("%s failed with rc=%d on %s", cmd, d.op_status(node), node)
    return all(d.is_ok(node) for node in nodes)

# vim:ts=4:sw=4:et:



import logging
import typing
import re
import os

from crmsh import sbd
from crmsh import watchdog
from crmsh import command
from crmsh import utils
from crmsh import bootstrap
from crmsh import completers
from crmsh import sh
from crmsh import xmlutil
from crmsh import constants
from crmsh import cibquery
from crmsh.service_manager import ServiceManager


logger = logging.getLogger(__name__)


def sbd_device_completer(completed_list: typing.List[str]) -> typing.List[str]:
    '''
    Completion for sbd device command
    '''
    if not sbd.SBDUtils.is_using_disk_based_sbd():
        return []
    if len(completed_list) == 2:
        return ["add", "remove"]
    if len(completed_list) > 2 and completed_list[1] != "remove":
        return []

    # completer for sbd device remove
    dev_list = sbd.SBDUtils.get_sbd_device_from_config()
    not_complete_list = [dev for dev in dev_list if dev not in completed_list[2:]]
    # not allow to remove the last device
    if len(not_complete_list) == 1:
        return []
    return not_complete_list


def sbd_configure_completer(completed_list: typing.List[str]) -> typing.List[str]:
    '''
    completion for sbd configure command
    '''
    service_manager = ServiceManager()
    pcmk_is_active = service_manager.service_is_active(constants.PCMK_SERVICE)
    sbd_is_active = service_manager.service_is_active(constants.SBD_SERVICE)
    if not pcmk_is_active or not sbd_is_active:
        return []

    is_diskbased = sbd.SBDUtils.is_using_disk_based_sbd()
    is_diskless = sbd.SBDUtils.is_using_diskless_sbd()
    show_types, timeout_types = (), ()
    if is_diskbased:
        show_types = SBD.SHOW_TYPES
        timeout_types = SBD.TIMEOUT_TYPES
    elif is_diskless:
        show_types = SBD.DISKLESS_SHOW_TYPES
        timeout_types = SBD.DISKLESS_TIMEOUT_TYPES

    if completed_list[1] == "show":
        if len(completed_list) == 3:
            return [t for t in show_types if t not in completed_list]
        else:
            return []

    parameters_pool = []
    if completed_list[1] == '':
        parameters_pool = ["show"]
    parameters_pool.extend([f"{t}-timeout=" for t in timeout_types])
    parameters_pool.append("watchdog-device=")
    parameters_pool = [
        p
        for p in parameters_pool
        if not any(c.startswith(p) for c in completed_list)
    ]
    return parameters_pool


class SBD(command.UI):
    '''
    Class for sbd sub-level

    Includes commands:
    - sbd configure
    - sbd device
    - sbd status
    - sbd purge
    '''
    name = "sbd"
    TIMEOUT_TYPES = ("watchdog", "allocate", "loop", "msgwait", "crashdump-watchdog")
    DISKLESS_TIMEOUT_TYPES = ("watchdog", "crashdump-watchdog")
    SHOW_TYPES = ("disk_metadata", "sysconfig", "property")
    DISKLESS_SHOW_TYPES = ("sysconfig", "property")
    PCMK_ATTRS = (
        "have-watchdog",
        "stonith-timeout",
        "stonith-enabled"
    )
    PCMK_ATTRS_DISKLESS = ('stonith-watchdog-timeout',)
    PARSE_RE = re.compile(
        # To extract key, suffix and value from these possible arguments:
        # watchdog-timeout=30
        # crashdump-watchdog-timeout=120
        # watchdog-device=/dev/watchdog
        r'([\w-]+)-([\w]+)=([\w/]+)'
    )
    # re pattern to match "-C <number>" or "-C <number> -Z"
    SBD_OPTS_RE = r'-C\s+\d+(\s+-Z)?'

    class SyntaxError(Exception):
        pass

    def __init__(self):
        self.device_list_from_config: list[str] = None
        self.device_meta_dict_runtime: dict[str, int] = None
        self.watchdog_timeout_from_config: int = None
        self.crashdump_watchdog_timeout_from_config: int = None
        self.watchdog_device_from_config: str = None
        self.service_manager: ServiceManager = None
        self.cluster_shell: sh.cluster_shell = None
        self.cluster_nodes: list[str] = None

        command.UI.__init__(self)

    def _load_attributes(self):
        if not os.path.isfile(sbd.SBDManager.SYSCONFIG_SBD):
            logger.error("SBD configuration file %s not found", sbd.SBDManager.SYSCONFIG_SBD)
            raise utils.TerminateSubCommand
        self.device_list_from_config = sbd.SBDUtils.get_sbd_device_from_config()
        self.device_meta_dict_runtime = {}
        if self.device_list_from_config:
            self.device_meta_dict_runtime = sbd.SBDUtils.get_sbd_device_metadata(self.device_list_from_config[0], timeout_only=True)
        try:
            self.watchdog_timeout_from_config = sbd.SBDTimeout.get_sbd_watchdog_timeout()
        except Exception:
            self.watchdog_timeout_from_config = None
        self.watchdog_device_from_config = watchdog.Watchdog.get_watchdog_device_from_sbd_config()
        self.crashdump_watchdog_timeout_from_config = sbd.SBDUtils.get_crashdump_watchdog_timeout()

        self.service_manager = ServiceManager()
        self.cluster_shell = sh.cluster_shell()
        self.cluster_nodes = utils.list_cluster_nodes() or [utils.this_node()]
        self.cluster_nodes = utils.get_reachable_node_list(self.cluster_nodes)

    def requires(self) -> bool:
        '''
        Requirements check when entering sbd sub-level
        '''
        if not utils.package_is_installed("sbd"):
            logger.error('%s', sbd.SBDManager.SBD_NOT_INSTALLED_MSG)
            return False
        return True

    def _service_is_active(self, service: str) -> bool:
        if not self.service_manager.service_is_active(service):
            logger.error("%s is not active", service)
            return False
        return True

    @property
    def configure_usage(self) -> str:
        '''
        Build usage string for sbd configure command,
        including disk-based and diskless sbd cases
        '''
        timeout_types, show_types = (), ()
        if sbd.SBDUtils.is_using_disk_based_sbd():
            timeout_types, show_types = self.TIMEOUT_TYPES, self.SHOW_TYPES
        elif sbd.SBDUtils.is_using_diskless_sbd():
            timeout_types, show_types = self.DISKLESS_TIMEOUT_TYPES, self.DISKLESS_SHOW_TYPES
        else:
            return ""

        timeout_usage_str = " ".join([f"[{t}-timeout=<integer>]" for t in timeout_types])
        show_usage = f"crm sbd configure show [{'|'.join(show_types)}]"
        return f"Usage:\n{show_usage}\ncrm sbd configure {timeout_usage_str} [watchdog-device=<device>]\n"

    def _show_sysconfig(self) -> None:
        '''
        Show pure content of /etc/sysconfig/sbd
        '''
        with open(sbd.SBDManager.SYSCONFIG_SBD) as f:
            content_list = [
                line.strip()
                for line in f.readlines()
                if not line.startswith("#") and line.strip()
            ]
        if content_list:
            logger.info("crm sbd configure show sysconfig")
        for line in content_list:
            if line.startswith("SBD_WATCHDOG_TIMEOUT") and bool(self.device_list_from_config):
                continue
            print(line)

    def _show_disk_metadata(self) -> None:
        '''
        Show sbd disk metadata for each configured device
        '''
        if self.device_list_from_config:
            logger.info("crm sbd configure show disk_metadata")
        for dev in self.device_list_from_config:
            print(self.cluster_shell.get_stdout_or_raise_error(f"sbd -d {dev} dump"))
            print()

    def _show_property(self) -> None:
        '''
        Show sbd-related properties from cluster and systemd
        '''
        out = self.cluster_shell.get_stdout_or_raise_error("crm configure show")

        logger.info("crm sbd configure show property")
        if self.device_list_from_config:
            attrs = self.PCMK_ATTRS
        else:
            attrs = self.PCMK_ATTRS + self.PCMK_ATTRS_DISKLESS
        regex = f"({'|'.join(attrs)})=(\\S+)"
        matches = re.findall(regex, out)
        for match in matches:
            print(f"{match[0]}={match[1]}")

        cmd = "crm configure show related:fence_sbd"
        out = self.cluster_shell.get_stdout_or_raise_error(cmd)
        if out:
            print()
            logger.info('%s', cmd)
            print(out)

        print()
        logger.info('%s', sbd.SBDTimeout.SHOW_SBD_START_TIMEOUT_CMD)
        systemd_start_timeout = sbd.SBDTimeout.get_sbd_systemd_start_timeout()
        print(f"TimeoutStartUSec={systemd_start_timeout}")

    def _configure_show(self, args) -> None:
        if len(args) > 2:
            raise self.SyntaxError("Invalid argument")
        elif len(args) == 2:
            match args[1]:
                case "disk_metadata":
                    self._show_disk_metadata()
                case "sysconfig":
                    self._show_sysconfig()
                case "property":
                    self._show_property()
                case _:
                    raise self.SyntaxError(f"Unknown argument: {args[1]}")
        else:
            self._show_disk_metadata()
            if self.device_list_from_config:
                print()
            self._show_sysconfig()
            print()
            self._show_property()

    def _parse_args(self, args: tuple[str, ...]) -> dict[str, int|str]:
        '''
        Parse arguments and verify them
        '''
        parameter_dict = {}

        for arg in args:
            match = self.PARSE_RE.match(arg)
            if not match:
                raise self.SyntaxError(f"Invalid argument: {arg}")
            key, suffix, value = match.groups()
            # timeout related parameters
            if key in self.TIMEOUT_TYPES and suffix and suffix == "timeout":
                if not value.isdigit():
                    raise self.SyntaxError(f"Invalid timeout value: {value}")
                parameter_dict[key] = int(value)
            # watchdog device parameter
            elif key == "watchdog" and suffix == "device":
                parameter_dict["watchdog-device"] = value
            else:
                raise self.SyntaxError(f"Unknown argument: {arg}")

        watchdog_device = parameter_dict.get("watchdog-device")
        parameter_dict["watchdog-device"] = watchdog.Watchdog.get_watchdog_device(watchdog_device)

        logger.debug("Parsed arguments: %s", parameter_dict)
        return parameter_dict

    @staticmethod
    def _adjust_timeout_dict(timeout_dict: dict) -> dict:
        watchdog_timeout = timeout_dict.get("watchdog")
        msgwait_timeout = timeout_dict.get("msgwait")
        if watchdog_timeout and msgwait_timeout and msgwait_timeout < 2*watchdog_timeout:
            logger.warning("It's recommended to set msgwait timeout >= 2*watchdog timeout")
            return timeout_dict
        if watchdog_timeout and not msgwait_timeout:
            timeout_dict["msgwait"] = 2*watchdog_timeout
            logger.info("No 'msgwait-timeout=' specified in the command, use 2*watchdog timeout: %s", 2*watchdog_timeout)
            return timeout_dict
        if msgwait_timeout and not watchdog_timeout:
            watchdog_timeout = msgwait_timeout//2
            timeout_dict["watchdog"] = watchdog_timeout
            logger.info("No 'watchdog-timeout=' specified in the command, use msgwait timeout/2: %s", watchdog_timeout)
            return timeout_dict
        return timeout_dict

    def _set_crashdump_option(self, delete=False):
        '''
        Set crashdump option for fence_sbd resource
        '''
        cib = xmlutil.text2elem(self.cluster_shell.get_stdout_or_raise_error('crm configure show xml'))
        ra = cibquery.ResourceAgent("stonith", "", "fence_sbd")
        res_id_list = cibquery.get_primitives_with_ra(cib, ra)
        if not res_id_list:
            if delete:
                return
            logger.error("No fence_sbd resource found")
            raise utils.TerminateSubCommand

        crashdump_value = cibquery.get_parameter_value(cib, res_id_list[0], "crashdump")
        cmd = ""
        if utils.is_boolean_false(crashdump_value):
            if delete:
                return
            cmd = f"crm resource param {res_id_list[0]} set crashdump 1"
            logger.info("Set crashdump option for fence_sbd resource")
        elif delete:
            cmd = f"crm resource param {res_id_list[0]} delete crashdump"
            logger.info("Delete crashdump option for fence_sbd resource")
        if cmd:
            self.cluster_shell.get_stdout_or_raise_error(cmd)

    def _set_crashdump_in_sysconfig(self, crashdump_watchdog_timeout=None, restore=False, diskless=False) -> dict:
        update_dict = {}
        sbd_timeout_action_for_crashdump = "flush,crashdump"
        comment_action_line = f"sed -i '/^SBD_TIMEOUT_ACTION/s/^/#__sbd_crashdump_backup__ /' {sbd.SBDManager.SYSCONFIG_SBD}"
        add_action_line = f"sed -i '/^#__sbd_crashdump_backup__/a SBD_TIMEOUT_ACTION={sbd_timeout_action_for_crashdump}' {sbd.SBDManager.SYSCONFIG_SBD}"
        comment_out_action_line = f"sed -i 's/^#__sbd_crashdump_backup__ SBD_TIMEOUT_ACTION/SBD_TIMEOUT_ACTION/' {sbd.SBDManager.SYSCONFIG_SBD}"
        delete_action_line = f"sed -i '/^SBD_TIMEOUT_ACTION/d' {sbd.SBDManager.SYSCONFIG_SBD}"

        sbd_timeout_action_configured = sbd.SBDUtils.get_sbd_value_from_config("SBD_TIMEOUT_ACTION")
        if restore:
            if sbd_timeout_action_configured and sbd_timeout_action_configured == sbd_timeout_action_for_crashdump:
                cmd_delete_and_comment_out = f"{delete_action_line} && {comment_out_action_line}"
                logger.info("Delete SBD_TIMEOUT_ACTION: %s and restore original value", sbd_timeout_action_for_crashdump)
                self.cluster_shell.get_stdout_or_raise_error(cmd_delete_and_comment_out)

            sbd_opts = sbd.SBDUtils.get_sbd_value_from_config("SBD_OPTS")
            if sbd_opts and re.search(self.SBD_OPTS_RE, sbd_opts):
                sbd_opts = re.sub(self.SBD_OPTS_RE, '', sbd_opts)
                update_dict["SBD_OPTS"] = ' '.join(sbd_opts.split())

        elif crashdump_watchdog_timeout:
            if not sbd_timeout_action_configured:
                update_dict["SBD_TIMEOUT_ACTION"] = sbd_timeout_action_for_crashdump
            elif sbd_timeout_action_configured != sbd_timeout_action_for_crashdump:
                cmd_comment_and_add = f"{comment_action_line} && {add_action_line}"
                self.cluster_shell.get_stdout_or_raise_error(cmd_comment_and_add)
                logger.info("Update SBD_TIMEOUT_ACTION in %s: %s", sbd.SBDManager.SYSCONFIG_SBD, sbd_timeout_action_for_crashdump)

            value_for_diskless = " -Z" if diskless else ""
            value_for_sbd_opts = f"-C {crashdump_watchdog_timeout}{value_for_diskless}"
            sbd_opts = sbd.SBDUtils.get_sbd_value_from_config("SBD_OPTS")
            if sbd_opts:
                sbd_opts = re.sub(self.SBD_OPTS_RE, '', sbd_opts)
            update_dict["SBD_OPTS"] = f"{' '.join(sbd_opts.split())} {value_for_sbd_opts}" if sbd_opts else value_for_sbd_opts

        return update_dict

    def _check_kdump_service(self):
        no_kdump = False
        for node in self.cluster_nodes:
            if not self.service_manager.service_is_active("kdump.service", node):
                logger.warning("Kdump service is not active on %s", node)
                no_kdump = True
        if no_kdump:
            logger.warning("Kdump service is required for crashdump")

    def _should_configure_crashdump(
            self,
            crashdump_watchdog_timeout,
            watchdog_timeout,
            diskless=False
        ) -> bool:
        if not crashdump_watchdog_timeout and not self.crashdump_watchdog_timeout_from_config:
            return False
        ct_updated = crashdump_watchdog_timeout and \
                crashdump_watchdog_timeout != self.crashdump_watchdog_timeout_from_config
        watchdog_timeout_configured = self.watchdog_timeout_from_config if diskless \
                else self.device_meta_dict_runtime.get("watchdog")
        wt_updated = watchdog_timeout and watchdog_timeout != watchdog_timeout_configured
        return ct_updated or wt_updated

    def _configure_diskbase(self, parameter_dict: dict):
        '''
        Configure disk-based SBD based on input parameters and runtime config
        '''
        update_dict = {}
        watchdog_device = parameter_dict.get("watchdog-device")
        if watchdog_device != self.watchdog_device_from_config:
            update_dict["SBD_WATCHDOG_DEV"] = watchdog_device

        timeout_dict = {
            k: v for k, v in parameter_dict.items()
            if k in self.TIMEOUT_TYPES and k != "crashdump-watchdog"
        }
        timeout_dict = self._adjust_timeout_dict(timeout_dict)
        # merge runtime timeout dict into parameter timeout dict without overwriting
        timeout_dict = {**self.device_meta_dict_runtime, **timeout_dict}

        crashdump_watchdog_timeout = parameter_dict.get("crashdump-watchdog", self.crashdump_watchdog_timeout_from_config)
        if self._should_configure_crashdump(crashdump_watchdog_timeout, timeout_dict.get("watchdog")):
            self._check_kdump_service()
            self._set_crashdump_option()
            timeout_dict["msgwait"] = 2*timeout_dict["watchdog"] + crashdump_watchdog_timeout
            logger.info("Set msgwait-timeout to 2*watchdog-timeout + crashdump-watchdog-timeout: %s", timeout_dict["msgwait"])
            result_dict = self._set_crashdump_in_sysconfig(crashdump_watchdog_timeout)
            update_dict = {**update_dict, **result_dict}

        if timeout_dict == self.device_meta_dict_runtime and not update_dict:
            logger.info("No change in SBD configuration")
            return

        sbd_manager = sbd.SBDManager(
            device_list_to_init=self.device_list_from_config,
            timeout_dict=timeout_dict,
            update_dict=update_dict
        )
        sbd_manager.init_and_deploy_sbd()
        
    def _configure_diskless(self, parameter_dict: dict):
        '''
        Configure diskless SBD based on input parameters and runtime config
        '''
        update_dict = {}
        timeout_dict = {}

        watchdog_timeout = parameter_dict.get("watchdog")
        if watchdog_timeout and watchdog_timeout != self.watchdog_timeout_from_config:
            update_dict["SBD_WATCHDOG_TIMEOUT"] = str(watchdog_timeout)
        watchdog_device = parameter_dict.get("watchdog-device")
        if watchdog_device != self.watchdog_device_from_config:
            update_dict["SBD_WATCHDOG_DEV"] = watchdog_device

        crashdump_watchdog_timeout = parameter_dict.get("crashdump-watchdog", self.crashdump_watchdog_timeout_from_config)
        if self._should_configure_crashdump(crashdump_watchdog_timeout, watchdog_timeout, diskless=True):
            self._check_kdump_service()
            result_dict = self._set_crashdump_in_sysconfig(crashdump_watchdog_timeout, diskless=True)
            update_dict = {**update_dict, **result_dict}
            sbd_watchdog_timeout = watchdog_timeout or self.watchdog_timeout_from_config
            stonith_watchdog_timeout = sbd_watchdog_timeout + crashdump_watchdog_timeout
            logger.info("Set stonith-watchdog-timeout to SBD_WATCHDOG_TIMEOUT + crashdump-watchdog-timeout: %s", stonith_watchdog_timeout)
            timeout_dict["stonith-watchdog"] = stonith_watchdog_timeout
        if not update_dict:
            logger.info("No change in SBD configuration")
            return

        sbd_manager = sbd.SBDManager(
            timeout_dict=timeout_dict,
            update_dict=update_dict,
            diskless_sbd=True
        )
        sbd_manager.init_and_deploy_sbd()

    def _device_add(self, devices_to_add: typing.List[str]):
        '''
        Implement sbd device add command, add devices to sbd configuration
        '''
        all_device_list = self.device_list_from_config + devices_to_add
        sbd.SBDUtils.verify_sbd_device(all_device_list)

        devices_to_init, _ = sbd.SBDUtils.handle_input_sbd_devices(
            devices_to_add,
            self.device_list_from_config
        )

        logger.info("Append devices: %s", ';'.join(devices_to_add))
        update_dict = {"SBD_DEVICE": ";".join(all_device_list)}
        sbd_manager = sbd.SBDManager(
            device_list_to_init=devices_to_init,
            update_dict=update_dict,
            timeout_dict=self.device_meta_dict_runtime
        )
        sbd_manager.init_and_deploy_sbd()

    def _device_remove(self, devices_to_remove: typing.List[str]):
        '''
        Implement sbd device remove command, remove devices from sbd configuration
        '''
        for dev in devices_to_remove:
            if dev not in self.device_list_from_config:
                raise self.SyntaxError(f"Device {dev} is not in config")
        # To keep the order of devices during removal
        left_device_list = [dev for dev in self.device_list_from_config if dev not in devices_to_remove]
        if len(left_device_list) == 0:
            raise self.SyntaxError("Not allowed to remove all devices")

        logger.info("Remove devices: %s", ';'.join(devices_to_remove))
        update_dict = {"SBD_DEVICE": ";".join(left_device_list)}
        sbd.SBDManager.update_sbd_configuration(update_dict)
        sbd.SBDManager.restart_cluster_if_possible()

    @command.completers_repeating(sbd_device_completer)
    def do_device(self, context, *args) -> bool:
        '''
        Implement sbd device command
        '''
        self._load_attributes()
        if not self._service_is_active(constants.PCMK_SERVICE):
            return False
        if not sbd.SBDUtils.is_using_disk_based_sbd():
            logger.error("Only works for disk-based SBD")
            logger.info("Please use 'crm cluster init sbd -s <dev1> [-s <dev2> [-s <dev3>]]' to configure the disk-based SBD first")
            return False

        try:
            if not args:
                raise self.SyntaxError("No argument")
            if args[0] not in ("add", "remove"):
                raise self.SyntaxError(f"Invalid argument: {args[0]}")
            if len(args) < 2:
                raise self.SyntaxError("No device specified")

            utils.check_all_nodes_reachable("configuring SBD device")

            logger.info("Configured sbd devices: %s", ';'.join(self.device_list_from_config))
            if len(args) == 2 and ";" in args[1]:
                device_list_from_args = args[1].split(";")
            else:
                device_list_from_args = list(args[1:])
            match args[0]:
                case "add":
                    self._device_add(device_list_from_args)
                case "remove":
                    self._device_remove(device_list_from_args)
            return True

        except self.SyntaxError as e:
            logger.error('%s', str(e))
            logger.info("Usage: crm sbd device <add|remove> <device>...")
            return False

    @command.completers_repeating(sbd_configure_completer)
    def do_configure(self, context, *args) -> bool:
        '''
        Implement sbd configure command
        '''
        try:
            self._load_attributes()
            if not args:
                raise self.SyntaxError("No argument")
            if args[0] == "show":
                self._configure_show(args)
                return True
            for service in (constants.PCMK_SERVICE, constants.SBD_SERVICE):
                if not self._service_is_active(service):
                    return False

            utils.check_all_nodes_reachable("configuring SBD")

            parameter_dict = self._parse_args(args)
            if sbd.SBDUtils.is_using_disk_based_sbd():
                self._configure_diskbase(parameter_dict)
            elif sbd.SBDUtils.is_using_diskless_sbd():
                self._configure_diskless(parameter_dict)
            return True

        except self.SyntaxError as e:
            logger.error('%s', str(e))
            usage = self.configure_usage
            if usage:
                print(usage)
            return False

    @command.completers(completers.choice(['crashdump']))
    def do_purge(self, context, *args) -> bool:
        '''
        Implement sbd purge command
        '''
        self._load_attributes()
        if not self._service_is_active(constants.SBD_SERVICE):
            return False

        utils.check_all_nodes_reachable("purging SBD")

        if args and args[0] == "crashdump":
            self._set_crashdump_option(delete=True)
            update_dict = self._set_crashdump_in_sysconfig(restore=True)
            if update_dict:
                sbd.SBDManager.update_sbd_configuration(update_dict)
                sbd.SBDManager.restart_cluster_if_possible()
            return True

        sbd.purge_sbd_from_cluster()
        sbd.SBDManager.restart_cluster_if_possible()
        return True

    def _print_sbd_type(self):
        if not self.service_manager.service_is_active(constants.SBD_SERVICE):
            return
        print("# Type of SBD:")
        if self.device_list_from_config:
            print("Disk-based SBD configured")
        else:
            print("Diskless SBD configured")
        print()

    def _print_sbd_status(self):
        padding = 2
        status_len = 8
        max_node_len = max(len(node) for node in self.cluster_nodes) + padding

        print(f"# Status of {constants.SBD_SERVICE}:")
        print(f"{'Node':<{max_node_len}}|{'Active':<{status_len}}|{'Enabled':<{status_len}}|Since")
        for node in self.cluster_nodes:
            is_active = self.service_manager.service_is_active(constants.SBD_SERVICE, node)
            is_active_str = "YES" if is_active else "NO"
            is_enabled = self.service_manager.service_is_enabled(constants.SBD_SERVICE, node)
            is_enabled_str = "YES" if is_enabled else "NO"
            systemd_property = "ActiveEnterTimestamp" if is_active else "ActiveExitTimestamp"
            since_str_prefix = "active since" if is_active else "disactive since"
            systemctl_show_cmd = f"systemctl show {constants.SBD_SERVICE} --property={systemd_property} --value"
            since = self.cluster_shell.get_stdout_or_raise_error(systemctl_show_cmd, node) or "N/A"
            print(f"{node:<{max_node_len}}|{is_active_str:<{status_len}}|{is_enabled_str:<{status_len}}|{since_str_prefix}: {since}")
        print()

    def _print_watchdog_info(self):
        padding = 2
        max_node_len = max(len(node) for node in self.cluster_nodes) + padding
        watchdog_sbd_re = r"\[[0-9]+\] (/dev/.*)\nIdentity: Busy: .*sbd.*\nDriver: (.*)"
        device_list, driver_list, kernel_timeout_list = [], [], []

        for node in self.cluster_nodes:
            out = self.cluster_shell.get_stdout_or_raise_error("sbd query-watchdog", node)
            res = re.search(watchdog_sbd_re, out)
            if res:
                device, driver = res.groups()
                kernel_timeout = self.cluster_shell.get_stdout_or_raise_error("cat /proc/sys/kernel/watchdog_thresh", node)
            else:
                device, driver, kernel_timeout = "N/A", "N/A", "N/A"
            device_list.append(device)
            driver_list.append(driver)
            kernel_timeout_list.append(kernel_timeout)

        print("# Watchdog info:")
        max_dev_len = max(len(dev) for dev in device_list+["Device"]) + padding
        max_driver_len = max(len(driver) for driver in driver_list+["Driver"]) + padding
        print(f"{'Node':<{max_node_len}}|{'Device':<{max_dev_len}}|{'Driver':<{max_driver_len}}|Kernel Timeout")
        for i, node in enumerate(self.cluster_nodes):
            print(f"{node:<{max_node_len}}|{device_list[i]:<{max_dev_len}}|{driver_list[i]:<{max_driver_len}}|{kernel_timeout_list[i]}")
        print()

    def _print_sbd_agent_status(self):
        for node in self.cluster_nodes:
            crm_mon_xml_parser = xmlutil.CrmMonXmlParser(node)
            if crm_mon_xml_parser.is_resource_configured(sbd.SBDManager.SBD_RA):
                print("# Status of fence_sbd:")
                sbd_id_list = crm_mon_xml_parser.get_resource_id_list_via_type(sbd.SBDManager.SBD_RA)
                for sbd_id in sbd_id_list:
                    rc, output = self.cluster_shell.get_rc_output_without_input(node, f"crm resource status {sbd_id}")
                    if output:
                        print(output)
                return

    def _print_sbd_cgroup_status(self):
        scripts_in_shell = '''#!/bin/bash
cgroup_procs_file="/sys/fs/cgroup/system.slice/sbd.service/cgroup.procs"
if [ ! -f "$cgroup_procs_file" ]; then
    exit
fi
pids=$(cat "$cgroup_procs_file")
for pid in $pids; do
    cmdline_file="/proc/$pid/cmdline"
    if [ -f "$cmdline_file" ]; then
        cmdline=$(tr '\0' ' ' < "$cmdline_file")
        if [[ "$cmdline" == *"slot:"* ]]; then
            echo "├─$pid \"$cmdline\""
        fi
    fi
done
        '''
        for node in self.cluster_nodes:
            out = self.cluster_shell.get_stdout_or_raise_error(scripts_in_shell, node)
            if out:
                print(f"# Status of the sbd disk watcher process on {node}:")
                print(out + "\n")

    def do_status(self, context) -> bool:
        '''
        Implement sbd status command
        '''
        self._load_attributes()
        self._print_sbd_type()
        self._print_sbd_status()
        self._print_sbd_cgroup_status()
        self._print_watchdog_info()
        self._print_sbd_agent_status()
        return True
